/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot_draw - handles the drawing of a plot
 */

#include <math.h>
#include <gtk/gtkmain.h>
#include <sys/time.h>

#include "plot_draw.xh"
#include "util.xh"
#include "gtkutil.xh"
#include "grid.xh"
#include "main.xh"

#if 0
#define pprintf printf
#else
#define pprintf(...)
#endif

static void plot_draw_legend(my_plot_t *plot, PangoFontDescription *font_desc, int w, int h);

static gint8 major_dashes[] = { 2, 2 };
static gint8 minor_dashes[] = { 1, 3 };

void plot_draw(my_plot_t *plot) {
   struct timeval start;
   gettimeofday(&start, 0);

   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);

   DA_ALLOCATION;

   // setup gc
   // ##TODO##
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;
      for(int i=0; i<line_count; ++i) {
	 if(line[i].line_gc) continue;
	 // printf("%d %s\n", i, line[i].color);

	 line[i].line_gc = x_gdk_gc_new_with_fg_color(plot, line[i].color);
	 line[i].horiz_marker = x_gdk_gc_new_with_fg_color(plot, line[i].color);

	 gint8 horiz_marker_dashes[] = { 2, 4 };
	 gdk_gc_set_dashes(line[i].horiz_marker, 0, horiz_marker_dashes, SZ(horiz_marker_dashes));
	 gdk_gc_set_line_attributes(line[i].horiz_marker, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
	 gdk_gc_set_function(line[i].horiz_marker, GDK_XOR);
      }
   }

   if(!plot->minor_gc) {
      plot->minor_gc = x_gdk_gc_new_with_fg_color(plot, plot->minor_color);
      plot->major_gc = x_gdk_gc_new_with_fg_color(plot, plot->major_color);
      plot->sep_gc = x_gdk_gc_new_with_fg_color(plot, plot->sep_color);
      plot->label_gc = x_gdk_gc_new_with_fg_color(plot, plot->label_color);
      plot->cursor_marker_gc = x_gdk_gc_new_with_fg_color(plot, plot->cursor_marker_color);

      gdk_gc_set_line_attributes(plot->minor_gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
      gdk_gc_set_line_attributes(plot->major_gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
      gdk_gc_set_line_attributes(plot->cursor_marker_gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);

      gdk_gc_set_function(plot->cursor_marker_gc, GDK_XOR);

      //	foreach my $family ($da->get_pango_context()->get_font_map()->list_families()) {
      //	    print %{$family}."\n";
      //	}

      plot->legend_bg_gc = x_gdk_gc_new_with_fg_color(plot, plot->legend_bg_color);
      plot->legend_border_gc = x_gdk_gc_new_with_fg_color(plot, plot->legend_border_color);
      plot->legend_text_gc = x_gdk_gc_new_with_fg_color(plot, plot->legend_text_color);
   }

   hide_markers(plot);
   plot->box = COORD_UNDEF;

   // my $queued = 0;

   //    gtk_main_iteration() while(gtk_events_pending());
   //    if($mytrans != plot->dsc_x->{num_changes}) {
   //	print "- $mytrans plot->dsc_x->{num_changes}\n";
   //#	    $da->queue_draw() unless($queued++);
   //	$conc--; return;
   //    }

   //############# grids

   // value grid
   double ymin = yoff;
   double ymax = yoff + ylength;

   pprintf("y %lf %lf\n", ymin, ymax);

   grid_t hgrid = calc_value_grid(ymin, ymax, h / 50, h);

   pprintf("%lf %lf %lf %lf %lf %lf %d\n", hgrid.dmin, hgrid.d, hgrid.dmax, hgrid.dvmin, hgrid.dv, hgrid.dvmax, hgrid.minpermaj);

   int doff;
   doff = (int)(xoff * w / xlength);
   gdk_gc_set_dashes(plot->major_gc, doff, major_dashes, SZ(major_dashes));
   gdk_gc_set_dashes(plot->minor_gc, doff, minor_dashes, SZ(minor_dashes));

   for(double yo = hgrid.dmin, yvo = hgrid.dvmin; yo < hgrid.dmax + hgrid.d; yo += hgrid.d, yvo += hgrid.dv) {
      if(yo < 0) continue;
      int y = h - (int)yo;
      char tmp[30];
      snprintf(tmp, sizeof(tmp), "%lf", yvo); // TODO conf decimal count
      PangoLayout *layout = pango_layout_new(plot->pangoctx);
      pango_layout_set_text(layout, tmp, -1);
      pango_layout_set_font_description(layout, plot->pangofont);
      gdk_draw_layout(plot->drawable, plot->label_gc, 0, y, layout);
      gdk_draw_line(plot->drawable, ((int)(round(yvo / hgrid.dv))) % hgrid.minpermaj == 0 ? plot->major_gc : plot->minor_gc, 0, y, w, y);
   }

   // time grid
   double xmin = xoff;
   double xmax = xoff + xlength;

   pprintf("x %lf %lf\n", xmin, xmax);

   grid_t vgrid = calc_time_grid(xmin, xmax, w / (plot->time_stamp_width * (1.3 / 2)), w);

   pprintf("%lf %lf %lf %lf %lf %lf %d\n", vgrid.dmin, vgrid.d, vgrid.dmax, vgrid.dvmin, vgrid.dv, vgrid.dvmax, vgrid.minpermaj);

   gdk_draw_line(plot->drawable, plot->sep_gc, 0, h, w, h);

   doff = -(int)(yoff * h / ylength);
   gdk_gc_set_dashes(plot->major_gc, doff, major_dashes, SZ(major_dashes));
   gdk_gc_set_dashes(plot->minor_gc, doff, minor_dashes, SZ(minor_dashes));

   int odd = ((int)round(vgrid.dvmin / vgrid.dv)) % 2;
   for(double xo = vgrid.dmin, xvo = vgrid.dvmin; xo < vgrid.dmax + vgrid.d; xo += vgrid.d, xvo += vgrid.dv) {
      odd = !odd;
      int x = (int)xo;
      PangoLayout *layout = pango_layout_new(plot->pangoctx);
      pango_layout_set_text(layout, format_time((my_time_t)round(xvo), 1), -1);
      pango_layout_set_font_description(layout, plot->pangofont);
      PangoRectangle ri;
      pango_layout_get_pixel_extents(layout, &ri, NULL);
      gdk_draw_layout(plot->drawable, plot->label_gc, x - ri.width/2, h + 1 + (odd ? 0 : (int)(ri.height * 1.3)), layout);
      gdk_draw_line(plot->drawable, ((int)round(xvo / vgrid.dv)) % vgrid.minpermaj == 0 ? plot->major_gc : plot->minor_gc, x, 0, x, h + (odd ? 0 : (int)(ri.height * 1.3 - 3 + ri.y)));
      // TODO 10 conf
   }

   my_time_t start_time = (my_time_t)xmin;
   my_time_t end_time = (my_time_t)xmax;

   //############ data

   // max points per native paint command
#define MAX_POINTS 1000

   double w_div_xlength = w / xlength;
   double xlength_div_w_ppx = xlength / (w * plot->samples_per_pixel);
   double h_div_vdiff = h / ylength;

   GdkRectangle clip = { 0, 0, w, h };

   // allocate points array fulfilling the maximum need
   int maxlines = 0;
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      maxlines = max(maxlines, plot->line_set[lsi].line_count);
   }
   GdkPoint *points = malloc_c(maxlines * MAX_POINTS * sizeof(GdkPoint));
   int *usedPoints = malloc_c(maxlines * sizeof(int));
   double *y = malloc_c(maxlines * sizeof(double));
   gboolean *gotundef = malloc_c(maxlines * sizeof(gboolean));

   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->line_set[lsi].dataset;
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      int start_index = my_time_arr_find(dataset->time, start_time, 0);
      int end_index = my_time_arr_find(dataset->time, end_time, 1);

      //	if(gtk_events_pending()) {
      //	    $da->queue_draw() unless($queued++);
      //
      //	    $conc--; return;
      //	}

      for(int li=0; li<line_count; ++li) {
	 if(!line[li].enabled) continue;
	 //#TODO## rectangle ref
	 gdk_gc_set_clip_rectangle(line[li].line_gc, &clip);
	 line[li].data = dataset->item[line[li].dataset_idx].data;
	 pprintf("[%d] = %lf\n", li, dataset->item[li].max);
      }

      double x = dataset->time[start_index] - xoff;

      // how many data entries can be skipped per row. Starts from 1, approaches correct value during rendering
      int skip_count = 1;

      bzero(usedPoints, line_count * sizeof(int));
      bzero(gotundef, line_count * sizeof(gboolean));

      pprintf("ii %d %d\n", start_index, end_index);

      for(int ti=start_index ; ti<=end_index; ) {
	 double start_x = x;

	 // ceil is more precise but 3-10% slower - the end result is we draw a little less points than requested
	 //double next_x = ceil(x * w_ppx_div_xlength) * xlength_div_w_ppx;
	 double next_x = x + xlength_div_w_ppx;

	 bzero(y, line_count * sizeof(double));
	 int start_ti = ti;
	 int span_count = 0;
	 while(1) {
	    for(int li=0; li<line_count; ++li) {
	       if(!line[li].enabled) continue;
	       double data = line[li].data[ti];
	       if(data > 84) {
		  pprintf("t%d i%d %lf\n", ti, li, data);
	       }
	       if(!__isnan(data)) {
		  switch(line[li].mode) {
		     case LINE_MODE_NORMAL:
			y[li] += data;
			break;
		     case LINE_MODE_DIFF:
			if(span_count == 0)
			   y[li] = data;
			break;
		  }
	       } else {
		  gotundef[li] = 1;
	       }
	    }
	    ti += skip_count;
	    span_count++;
	    if(ti > end_index) break;
	    x = dataset->time[ti] - xoff; // ##TODO## this sometimes gives undef value in subtract
	    if(x > next_x) break;
	 }

	 // adjust skip value
	 skip_count = span_count * skip_count / 4;
	 if(skip_count < 1) skip_count = 1;

	 if((dataset->time[start_ti] - dataset->time[ti]) / (start_ti - ti) > dataset->maxdiff) {
	    memset(gotundef, 1, line_count * sizeof(gboolean));
	 }

	 for(int li=0; li<line_count; ++li) {
	    if(!line[li].enabled) continue;
	    if(gotundef[li]) {
	       gotundef[li] = 0;
	       //pprintf("gun %d %d\n", li, usedPoints[li]);
	       if(usedPoints[li] > 1) {
		  pprintf("gu %d %d\n", li, usedPoints[li]);
		  gdk_draw_lines(plot->drawable, line[li].line_gc, &points[MAX_POINTS * li], usedPoints[li]);
	       }
	       usedPoints[li] = 0;
	    } else {
	       double yv;
	       switch(line[li].mode) {
		  case LINE_MODE_NORMAL:
		     yv = span_count > 1 ? y[li] / span_count : y[li];
		     break;
		  case LINE_MODE_DIFF: {
		     int end_ti = ti < end_index ? ti : end_index;
//		     yv = (line[li].data[end_ti] - y[li]) * TIME_MULTIPLIER / (dataset->time[end_ti] - dataset->time[start_ti]);
		     yv = (line[li].data[end_ti] - y[li]) * TIME_MULTIPLIER / (dataset->time[end_ti] - dataset->time[start_ti]) * (1 * TIME_MULTIPLIER);
		     break;
		     default:
			fprintf(stderr, "Bad mode %d in %d.%d\n", line[li].mode, lsi, li);
			return;
		  }
	       }

	       // calculate point coordinates
	       GdkPoint *point = &points[MAX_POINTS * li + usedPoints[li]++]; //##TODO##
	       point->x = start_x * w_div_xlength;
	       int ly = (ymax - yv) * h_div_vdiff;

	       // with extreme overdrives, these tend to go wrong (or so I guess)
	       if(ly > 32767) ly = 32767;
	       else if(ly < -32768) ly = -32768;
	       point->y = ly;

	       if(usedPoints[li] == MAX_POINTS) {
		  pprintf("gm %d %d\n", li, usedPoints[li]);
		  gdk_draw_lines(plot->drawable, line[li].line_gc, &points[MAX_POINTS * li], usedPoints[li]);
		  points[MAX_POINTS * li] = points[MAX_POINTS * li + usedPoints[li]-1];
		  usedPoints[li] = 1;
	       }
	    }
	 }
      }

      // plot final lines
      for(int li=0; li<line_count; ++li) {
	 if(!line[li].enabled) continue;
	 if(usedPoints[li] > 1) {
	    pprintf("f  %d %d\n", li, usedPoints[li]);
	    gdk_draw_lines(plot->drawable, line[li].line_gc, &points[MAX_POINTS * li], usedPoints[li]);
	 }
      }

      pprintf("skip_count = %d\n", skip_count);

   }

   free(points);
   free(usedPoints);
   free(y);
   free(gotundef);

   while(gtk_events_pending()) gtk_main_iteration();
   if(quit && gtk_main_level()) gtk_main_quit();

   plot_draw_legend(plot, plot->pangofont, w, h);

   struct timeval end;
   gettimeofday(&end, 0);

   plot->last_render_time = (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec) * 0.000001;

   while(gtk_events_pending()) gtk_main_iteration();
   if(quit && gtk_main_level()) gtk_main_quit();
}

static void plot_draw_legend(my_plot_t *plot, PangoFontDescription *font_desc, int w, int h) {

   // precalculate layouts to get max layout width
   int layout_max_size_x = 0;
   int layout_max_size_y = 0;
   int layout_min_ascent_y = 10;

   int list_count = 0;
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->line_set[lsi].dataset;
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      for(int li=0; li<line_count; ++li) {
	 if(!line[li].enabled && !plot->show_disabled) continue;

	 const char *linename = dataset->item[line[li].dataset_idx].name;
	 char *text = malloc((dataset->name ? strlen(dataset->name) : 7 + 1 + 10) + 2 + (linename ? strlen(linename) : 4 + 1 + 10) + (line[li].mode == LINE_MODE_DIFF ? 7 : 0) + 1);

	 if(plot->line_set_count > 1) {
	    if(dataset->name) {
	       strcpy(text, dataset->name);
	       strcat(text, ": ");
	    } else {
	       sprintf(text, "dataset %d: ", lsi + 1);
	    }
	 } else {
	    text[0] = '\0';
	 }

	 if(linename) {
	    strcat(text, linename);
	 } else {
	    sprintf(text + strlen(text), "item %d", li + 1);
	 }

	 switch(line[li].mode) {
	    case LINE_MODE_NORMAL:
	       break;

	    case LINE_MODE_DIFF:
	       strcat(text, " (diff)");
	       break;
	 }

	 PangoLayout *layout = pango_layout_new(plot->pangoctx);
	 pango_layout_set_text(layout, text, -1);
	 pango_layout_set_font_description(layout, font_desc);

	 // overstrike if not enabled
	 if(!line[li].enabled) {
	    PangoAttribute *pa_st = pango_attr_strikethrough_new(1);
	    pa_st->start_index = 0;
	    pa_st->end_index = strlen(text);
	    PangoAttrList *pal = pango_attr_list_new();
	    pango_attr_list_insert(pal, pa_st);
	    pango_layout_set_attributes(layout, pal);
	 }

	 PangoRectangle ri;
	 pango_layout_get_pixel_extents(layout, &ri, NULL);

	 layout_max_size_x = max(layout_max_size_x, ri.width);
	 layout_max_size_y = max(layout_max_size_y, ri.height);
	 layout_min_ascent_y = min(layout_min_ascent_y, ri.y);

/*
	 if(li >= 0) {
	    int lx, ly;
	    pango_layout_get_pixel_size(layout, &lx, &ly);
	    printf("h %d+%d %d\n", ri.y, ri.height, ly);
	 }
*/

	 free(text);

	 line[li].legend_layout = layout;

	 list_count++;
      }
   }

   layout_min_ascent_y--;
   if(layout_min_ascent_y < 0) layout_max_size_y -= layout_min_ascent_y;

   if(list_count == 0) return;

   double box_size_multiplier = 0.5;

   int legend_list_box_offset_x = 0;
   int legend_list_box_offset_y = (layout_max_size_y - (int)(layout_max_size_y * box_size_multiplier) + 1) / 2;
   int legend_list_box_size_x = (int)(layout_max_size_y * box_size_multiplier);
   int legend_list_box_size_y = (int)(layout_max_size_y * box_size_multiplier);
   int legend_list_box_text_spacing_x = 2 + legend_list_box_size_x / 2;

   // printf("%d %d %d\n", legend_list_box_offset_y, legend_list_box_size_y, layout_max_size_y);

   int legend_list_unit_size_x = layout_max_size_x + legend_list_box_offset_x + legend_list_box_size_x + legend_list_box_text_spacing_x;
   int legend_list_unit_size_y = layout_max_size_y;

   int legend_list_size_x = legend_list_unit_size_x;
   int legend_list_size_y = legend_list_unit_size_y * list_count;

   int legend_margin_x = 10;
   int legend_margin_y = 10;

   int legend_list_padding_x = 4 + legend_list_box_size_x / 2;
   int legend_list_padding_y = 5;

   int legend_corner_x = w - legend_margin_x - 2 * legend_list_padding_x - legend_list_size_x;
   int legend_corner_y = legend_margin_y;

   int legend_corner_x2 = legend_corner_x + 2 * legend_list_padding_x + legend_list_size_x;
   int legend_corner_y2 = legend_corner_y + 2 * legend_list_padding_y + legend_list_size_y;

   // draw box
   x_gdk_draw_rectangle(plot->drawable, plot->legend_bg_gc,     1, legend_corner_x+1, legend_corner_y+1, legend_corner_x2-1, legend_corner_y2-1);
   x_gdk_draw_rectangle(plot->drawable, plot->legend_border_gc, 0, legend_corner_x  , legend_corner_y  , legend_corner_x2  , legend_corner_y2  );

   plot->legend_x = legend_corner_x;
   plot->legend_y = legend_corner_y;
   plot->legend_x2 = legend_corner_x2;
   plot->legend_y2 = legend_corner_y2;

   int no = 0;

   // draw legend items
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      for(int li=0; li<line_count; ++li) {
	 if(!line[li].enabled && !plot->show_disabled) continue;

	 int legend_list_corner_x = legend_corner_x + legend_list_padding_x;
	 int legend_list_corner_y = legend_corner_y + legend_list_padding_y + legend_list_unit_size_y * no;

	 line[li].legend_x = legend_list_corner_x;
	 line[li].legend_y = legend_list_corner_y;
	 line[li].legend_x2 = legend_list_corner_x + legend_list_unit_size_x;
	 line[li].legend_y2 = legend_list_corner_y + legend_list_unit_size_y;

	 int legend_list_text_corner_x = legend_list_corner_x + legend_list_box_offset_x + legend_list_box_size_x + legend_list_box_text_spacing_x;
	 int legend_list_text_corner_y = legend_list_corner_y;

	 x_gdk_draw_rectangle(plot->drawable, line[li].enabled ? line[li].line_gc : plot->legend_border_gc, line[li].enabled,
			      legend_list_corner_x + legend_list_box_offset_x,
			      legend_list_corner_y + legend_list_box_offset_y,
			      legend_list_corner_x + legend_list_box_offset_x + legend_list_box_size_x - 1,
			      legend_list_corner_y + legend_list_box_offset_y + legend_list_box_size_y - 1);

	 /*
	 gdk_gc_set_function(plot->legend_border_gc, GDK_XOR);
	 x_gdk_draw_rectangle(plot->drawable, plot->legend_border_gc, 0,
			      legend_list_corner_x, legend_list_corner_y,
			      legend_list_corner_x + legend_list_unit_size_x - 1, legend_list_corner_y + legend_list_unit_size_y - 1);
	 gdk_gc_set_function(plot->legend_border_gc, GDK_SET);
	 */

	 gdk_draw_layout(plot->drawable, plot->legend_text_gc,
			 legend_list_text_corner_x,
			 legend_list_text_corner_y - layout_min_ascent_y, line[li].legend_layout);

/*
	 if(!line[li].enabled) {
	    printf("%d %d\n", lx, ly);
	    gdk_draw_line(plot->drawable, plot->legend_text_gc,
			  legend_list_text_corner_x, legend_list_text_corner_y + ly/2,
			  legend_list_text_corner_x + lx, legend_list_text_corner_y + ly/2);
	 }
*/

	 no++;
      }
   }

   //    last outer2 if(normal_expose);

   //	    if(mytrans != plot->dsc_x->{num_changes} + plot->dsc_y->{num_changes}) {
   //	    print "- mytrans plot->dsc_x->{num_changes}\n";
   //	    da->queue_draw() unless(queued++);
   //	    queued++;
   //	    plot->conc--; return;

   //####################
}
